"""
Indexing module
"""

import os.path
import sqlite3
import sys

import regex as re

from txtai.embeddings import Embeddings
from txtai.tokenizer import Tokenizer

from .models import Models

class Index(object):
    """
    Methods to build a new sentence embeddings index.
    """

    # Section query and filtering logic constants
    SECTION_FILTER = r"background|(?<!.*?results.*?)discussion|introduction|reference"
    SECTION_QUERY = "SELECT Id, Name, Text FROM sections WHERE (labels is null or labels NOT IN ('FRAGMENT', 'QUESTION'))"

    @staticmethod
    def stream(dbfile, maxsize):
        """
        Streams documents from an articles.sqlite file. This method is a generator and will yield a row at time.

        Args:
            dbfile: input SQLite file
            maxsize: maximum number of documents to process
        """

        # Connection to database file
        db = sqlite3.connect(dbfile)
        cur = db.cursor()

        # Select tagged sentences without a NLP label. NLP labels are set for non-informative sentences.
        query = Index.SECTION_QUERY + " AND tags is not null"

        if maxsize > 0:
            query += " AND article in (SELECT id FROM articles ORDER BY entry DESC LIMIT %d)" % maxsize

        # Run the query
        cur.execute(query)

        count = 0
        for row in cur:
            # Unpack row
            uid, name, text = row

            if not name or not re.search(Index.SECTION_FILTER, name.lower()):
                # Tokenize text
                tokens = Tokenizer.tokenize(text)

                document = (uid, tokens, None)

                count += 1
                if count % 1000 == 0:
                    print("Streamed %d documents" % (count), end="\r")

                # Skip documents with no tokens parsed
                if tokens:
                    yield document

        print("Iterated over %d total rows" % (count))

        # Free database resources
        db.close()

    @staticmethod
    def embeddings(dbfile, vectors, maxsize):
        """
        Builds a sentence embeddings index.

        Args:
            dbfile: input SQLite file
            vectors: vector path
            maxsize: maximum number of documents to process

        Returns:
            embeddings index
        """

        embeddings = Embeddings({"path": vectors,
                                 "scoring": "bm25",
                                 "pca": 3,
                                 "quantize": True})

        # Build scoring index if scoring method provided
        if embeddings.config.get("scoring"):
            embeddings.score(Index.stream(dbfile, maxsize))

        # Build embeddings index
        embeddings.index(Index.stream(dbfile, maxsize))

        return embeddings

    @staticmethod
    def run(path, vectors, maxsize=0):
        """
        Executes an index run.

        Args:
            path: model path, if None uses default path
            vectors: vector path, if None uses default path
            maxsize: maximum number of documents to process
        """

        # Default path if not provided
        if not path:
            path = Models.modelPath()

        dbfile = os.path.join(path, "articles.sqlite")

        # Default vectors
        if not vectors:
            vectors = Models.vectorPath("cord19-300d.magnitude")

        print("Building new model")
        embeddings = Index.embeddings(dbfile, vectors, maxsize)
        embeddings.save(path)

if __name__ == "__main__":
    Index.run(sys.argv[1] if len(sys.argv) > 1 else None,
              sys.argv[2] if len(sys.argv) > 2 else None,
              int(sys.argv[3]) if len(sys.argv) > 3 else 0)



"""
Report module
"""

import regex as re

from txtai.extractor import Extractor

from ..index import Index
from ..query import Query

class Report(object):
    """
    Methods to build reports from a series of queries
    """

    def __init__(self, embeddings, db, qa):
        """
        Creates a new report.

        Args:
            embeddings: embeddings index
            db: database connection
            qa: qa model path
        """

        # Store references to embeddings index and open database cursor
        self.embeddings = embeddings
        self.cur = db.cursor()

        # Column names
        self.names = []

        # Extractive question-answering model
        self.extractor = Extractor(self.embeddings, qa if qa else "NeuML/bert-small-cord19qa", False)

    def build(self, queries, topn, output):
        """
        Builds a report using a list of input queries

        Args:
            queries: queries to execute
            topn: number of documents to return
            output: output I/O object
        """

        # Default to 50 documents if not specified
        topn = topn if topn else 50

        for name, config in queries:
            query = config["query"]
            columns = config["columns"]

            # Write query string
            self.query(output, name, query)

            # Write separator
            self.separator(output)

            # Query for best matches
            results = Query.search(self.embeddings, self.cur, query, topn)

            # Generate highlights section
            self.section(output, "Highlights")

            # Generate highlights
            self.highlights(output, results, int(topn / 10))

            # Separator between highlights and articles
            self.separator(output)

            # Generate articles section
            self.section(output, "Articles")

            # Generate table headers
            self.headers([column["name"] for column in columns], output)

            # Generate table rows
            self.articles(output, topn, (name, query, columns), results)

            # Write section separator
            self.separator(output)

    def highlights(self, output, results, topn):
        """
        Builds a highlights section.

        Args:
            output: output file
            results: search results
            topn: number of results to return
        """

        # Extract top sections as highlights
        for highlight in Query.highlights(results, topn):
            # Get matching article
            uid = [article for _, _, article, text in results if text == highlight][0]
            self.cur.execute("SELECT Authors, Reference FROM articles WHERE id = ?", [uid])
            article = self.cur.fetchone()

            # Write out highlight row
            self.highlight(output, article, highlight)

    def articles(self, output, topn, metadata, results):
        """
        Builds an articles section.

        Args:
            output: output file
            topn: number of documents to return
            metadata: query metadata
            results: search results
        """

        # Unpack metadata
        _, query, _ = metadata

        # Retrieve list of documents
        documents = Query.all(self.cur) if query == "*" else Query.documents(results, topn)

        # Collect matching rows
        rows = []

        for uid in documents:
            # Get article metadata
            self.cur.execute("SELECT Published, Title, Reference, Publication, Source, Design, Size, Sample, Method, Entry " +
                             "FROM articles WHERE id = ?", [uid])
            article = self.cur.fetchone()

            # Calculate derived fields
            calculated = self.calculate(uid, metadata)

            # Builds a row for article
            rows.append(self.buildRow(article, documents[uid], calculated))

        # Print report by published desc
        for row in sorted(rows, key=lambda x: x["Date"], reverse=True):
            # Convert row dict to list
            row = [row[column] for column in self.names]

            # Write out row
            self.writeRow(output, row)

    def calculate(self, uid, metadata):
        """
        Builds a dict of calculated fields for a given document. This method calculates
        constant field columns and derived query columns. Derived query columns run through
        an embedding search and either run an additional QA query to extract a value or
        use the top n embedding search matches.

        Args:
            uid: article id
            metadata: query metadata

        Returns:
            {name: value} containing derived column values
        """

        # Parse column parameters
        fields, params = self.params(metadata)

        # Stores embedding query only and full QA column definitions
        queries, questions = [], []

        # Retrieve indexed document text for article
        sections = self.sections(uid)

        for name, query, question, snippet, _, _, matches in params:
            if matches:
                queries.append((name, query, matches))
            else:
                questions.append((name, query, question, snippet))

        # Only execute embeddings queries for columns with matches set
        for name, query, matches in queries:
            # Run query
            topn = self.extractor.query(sections, [query])[0]

            # Get topn text matches
            topn = [text for _, text, _ in topn][:matches]

            # Join results into String and return
            fields[name] = "\n\n".join([self.resolve(params, sections, uid, name, value) for value in topn])

        # Add extraction fields
        for name, value in self.extractor(sections, questions):
            # Resolvesthe full value based on column parameters
            fields[name] = self.resolve(params, sections, uid, name, value) if value else ""

        return fields

    def params(self, metadata):
        """
        Process and prepare parameters using input metadata.

        Args:
            metadata: query metadata

        Returns:
            fields, params - constant field values, query parameters for query columns
        """

        # Derived field values
        fields = {}

        # Query column parameters
        params = []

        # Unpack metadata
        _, _, columns = metadata

        for column in columns:
            # Constant column
            if "constant" in column:
                fields[column["name"]] = column["constant"]
            # Question-answer column
            elif "query" in column:
                # Query variable substitutions
                query = self.variables(column["query"], metadata)
                question = self.variables(column["question"], metadata) if "question" in column else query

                # Additional context parameters
                section = column["section"] if "section" in column else False
                surround = column["surround"] if "surround" in column else 0
                matches = column["matches"] if "matches" in column else 0
                snippet = column["snippet"] if "snippet" in column else False
                snippet = True if section or surround else snippet

                params.append((column["name"], query, question, snippet, section, surround, matches))

        return fields, params

    def variables(self, value, metadata):
        """
        Runs variable substitution for value.

        Args:
            value: input value
            metadata: query metadata

        Returns:
            value with variable substitution
        """

        name, query, _ = metadata

        # Cleanup name for queries
        name = name.replace("_", "").lower()
        query = query.lower()

        if value:
            value = value.replace("$NAME", name).replace("$QUERY", query)

        return value

    def sections(self, uid):
        """
        Retrieves all sections as list for article with given uid.

        Args:
            uid: article id

        Returns:
            list of section text elements
        """

        # Retrieve indexed document text for article
        self.cur.execute(Index.SECTION_QUERY + " AND article = ?", [uid])

        # Get list of document text sections
        sections = []
        for sid, name, text in self.cur.fetchall():
            if not name or not re.search(Index.SECTION_FILTER, name.lower()):
                sections.append((sid, text))

        return sections

    def resolve(self, params, sections, uid, name, value):
        """
        Fully resolves a value from an extractor call.

         - If section=True, this method pull the full section text
         - If surround is specified, this method will pull the surrounding text
         - Otherwise, the original value is returned

        Args:
            params: query parameters
            sections: section text
            uid: article id
            name: column name
            value: initial query value after running through extractor process

        Returns:
            resolved value
        """

        # Get all column parameters
        index = [params.index(x) for x in params if x[0] == name][0]
        _, _, _, _, section, surround, _ = params[index]

        if value:
            # Find matching section
            sid = [sid for sid, text in sections if value in text]

            if sid:
                sid = sid[0]

                if section:
                    # Get full text for matching subsection
                    value = self.subsection(uid, sid)
                elif surround:
                    value = self.surround(uid, sid, surround)

        return value

    def subsection(self, uid, sid):
        """
        Extracts all subsection text for columns with section=True.

        Args:
            uid: article id
            sid: section id

        Returns:
            full text for matching section
        """

        self.cur.execute("SELECT Text FROM sections WHERE article = ? AND name = (SELECT name FROM sections WHERE id = ?)", [uid, sid])
        return " ".join([x[0] for x in self.cur.fetchall()])

    def surround(self, uid, sid, size):
        """
        Extracts surrounding text for section with specified id.

        Args:
            uid: article id
            sid: section id
            size: number of surrounding lines to extract from each side

        Returns:
            matching text with surrounding context
        """

        self.cur.execute("SELECT Text FROM sections WHERE article = ? AND id in (SELECT id FROM sections WHERE id >= ? AND id <= ?) AND " + \
                         "name = (SELECT name FROM sections WHERE id = ?)", [uid, sid - size, sid + size, sid])

        return " ".join([x[0] for x in self.cur.fetchall()])

    def cleanup(self, outfile):
        """
        Allow freeing or cleaning up resources.

        Args:
            outfile: output file path
        """

    def query(self, output, task, query):
        """
        Writes query.

        Args:
            output: output file
            task: task name
            query: query string
        """

    def section(self, output, name):
        """
        Writes a section name

        Args:
            output: output file
            name: section name
        """

    def highlight(self, output, article, highlight):
        """
        Writes a highlight row

        Args:
            output: output file
            article: article reference
            highlight: highlight text
        """

    def headers(self, columns, output):
        """
        Writes table headers.

        Args:
            columns: column names
            output: output file
        """

    def buildRow(self, article, sections, calculated):
        """
        Converts a document to a table row.

        Args:
            article: article
            sections: text sections for article
            calculated: calculated fields
        """

    def writeRow(self, output, row):
        """
        Writes a table row.

        Args:
            output: output file
            row: output row
        """

    def separator(self, output):
        """
        Writes a separator between sections
        """



"""
Markdown report module
"""

from ..query import Query

from .common import Report

class Markdown(Report):
    """
    Report writer for Markdown.
    """

    def encode(self, url):
        """
        URL encodes parens as they cause issues with markdown links.

        Args:
            url: input url

        Returns:
            url with parens encoded
        """

        # Escape ()
        return url.replace("(", "%28").replace(")", "%29") if url else url

    def column(self, value):
        """
        Escapes invalid characters (| char) within a table column value.

        Args:
            value: input value

        Returns:
            value with | escaped
        """

        # Escape |
        return value.replace("|", "&#124;") if value else value

    def write(self, output, line):
        """
        Writes line to output file.

        Args:
            output: output file
            line: line to write
        """

        output.write("%s\n" % line)

    def query(self, output, task, query):
        self.write(output, "# %s" % query)

    def section(self, output, name):
        self.write(output, "#### %s<br/>" % name)

    def highlight(self, output, article, highlight):
        # Build citation link
        link = "[%s](%s)" % (Query.authors(article[0]) if article[0] else "Source", self.encode(article[1]))

        # Build highlight row with citation link
        self.write(output, "- %s %s<br/>" % (Query.text(highlight), link))

    def headers(self, columns, output):
        self.names = columns

        # Remove extended study columns
        for field in ["Journal", "Study Link", "Sample Text"]:
            if field in self.names:
                self.names.remove(field)

        # Write table header
        headers = "|".join(self.names)
        self.write(output, "|%s|" % headers)

        # Write markdown separator for headers
        headers = "|".join(["----"] * len(self.names))
        self.write(output, "|%s|" % headers)

    def buildRow(self, article, sections, calculated):
        columns = {}

        # Date
        columns["Date"] = Query.date(article[0]) if article[0] else ""

        # Title
        title = "[%s](%s)" % (article[1], self.encode(article[2]))

        # Append Publication if available. Assume preprint otherwise and show preprint source.
        title += "<br/>%s" % (article[3] if article[3] else article[4])

        # Source
        columns["Source"] = article[4]

        # Title + Publication if available
        columns["Study"] = title

        # Study Type
        columns["Study Type"] = Query.design(article[5])

        # Sample Size
        sample = Query.sample(article[6], article[7])
        columns["Sample Size"] = sample if sample else ""

        # Study Population
        columns["Study Population"] = Query.text(article[8]) if article[8] else ""

        # Top Matches
        columns["Matches"] = "<br/><br/>".join([Query.text(text) for _, text in sections]) if sections else ""

        # Entry Date
        columns["Entry"] = article[9] if article[9] else ""

        # Merge in calculated fields
        columns.update(calculated)

        # Escape | characters embedded within columns
        return {column: self.column(columns[column]) for column in columns}

    def writeRow(self, output, row):
        self.write(output, "|%s|" % "|".join(row))

    def separator(self, output):
        # Write section separator
        self.write(output, "")



"""
Highlights module
"""

import itertools

import networkx

from txtai.tokenizer import Tokenizer

class Highlights(object):
    """
    Methods to extract highlights from a list of text sections.
    """

    # Domain specific stop list
    STOP_WORDS = {"abstract", "al", "article", "arxiv", "author", "biorxiv", "copyright", "da", "dei", "del", "dell", "della",
                  "delle", "di", "doi", "et", "fig", "figure", "funder", "holder", "http", "https", "il", "la", "le",
                  "license", "medrxiv", "non", "org", "peer", "peer-reviewed", "permission", "preprint", "publication",
                  "pubmed", "reserved", "reviewed", "rights", "si", "una", "used", "using"}

    @staticmethod
    def build(sections, topn):
        """
        Extracts highlights from a list of sections. This method uses textrank to find sections with the highest
        importance across the input list. This method attempts to return important but unique results to limit
        repetitive statements.

        Args:
            sections: input sections
            topn: top n results to return

        Results:
            top n sections
        """

        results = []

        # Rank the text using textrank for importance within collection
        for uid, _ in Highlights.textrank(sections):
            # Lookup text and tokenize
            text = [text for u, text in sections if u == uid][0]
            tokens = Highlights.tokenize(text)

            # Compare text to existing results, look for highly unique results
            # This finds results that are important but not repetitive
            unique = all([Highlights.jaccardIndex(t, tokens) <= 0.2 for _, t in results])
            if unique:
                results.append((uid, tokens))

        uids = [uid for uid, _ in results][:topn]

        # Get related text for each match
        return [text for uid, text in sections if uid in uids]

    @staticmethod
    def textrank(sections):
        """
        Runs the textrank algorithm against the list of sections. Orders the list into descending order of importance
        given the list.

        Args:
            sections: list of sentences

        Returns:
            sorted list using the textrank algorithm
        """

        # Build the graph network
        graph = Highlights.buildGraph(sections)

        # Run pagerank
        rank = networkx.pagerank(graph, weight="weight")

        # Return items sorted by highest score first
        return sorted(list(rank.items()), key=lambda x: x[1], reverse=True)

    @staticmethod
    def buildGraph(nodes):
        """
        Builds a graph of nodes using input.

        Args:
            nodes: input graph nodes

        Returns:
            graph
        """

        graph = networkx.Graph()
        graph.add_nodes_from([uid for (uid, _) in nodes])

        # Tokenize nodes, store uid and tokens
        vectors = []
        for (uid, text) in nodes:
            # Custom tokenization that works best with textrank matching
            tokens = Highlights.tokenize(text)

            if len(tokens) >= 3:
                vectors.append((uid, tokens))

        pairs = list(itertools.combinations(vectors, 2))

        # add edges to the graph
        for pair in pairs:
            node1, tokens1 = pair[0]
            node2, tokens2 = pair[1]

            # Add a graph edge and compute the cosine similarity for the weight
            graph.add_edge(node1, node2, weight=Highlights.jaccardIndex(tokens1, tokens2))

        return graph

    @staticmethod
    def jaccardIndex(set1, set2):
        """
        Jaccard index calculation used for similarity.

        Args:
            set1: input 1
            set2: input 2

        Returns:
            jaccard index
        """

        n = len(set1.intersection(set2))
        return n / float(len(set1) + len(set2) - n) if n > 0 else 0

    @staticmethod
    def tokenize(text):
        """
        Tokenizes text into tokens, removes domain specific stop words.

        Args:
            text: input text

        Returns:
            tokens
        """

        # Remove additional stop words to improve highlighting results
        return {token for token in Tokenizer.tokenize(text) if token not in Highlights.STOP_WORDS}



"""
Query module
"""

import datetime
import re
import sys

import html2text
import mdv

from txtai.tokenizer import Tokenizer

from .highlights import Highlights
from .models import Models

class Query(object):
    """
    Methods to query an embeddings index.
    """

    @staticmethod
    def escape(text):
        """
        Escapes text to work around issues with mdv double escaping characters.

        Args:
            text: input text

        Returns:
            escaped text
        """

        text = text.replace("<", "¿")
        text = text.replace(">", "Ñ")
        text = text.replace("&", "ž")

        return text

    @staticmethod
    def unescape(text):
        """
        Un-escapes text to work around issues with mdv double escaping characters.

        Args:
            text: input text

        Returns:
            unescaped text
        """

        text = text.replace("¿", "<")
        text = text.replace("Ñ", ">")
        text = text.replace("ž", "&")

        return text

    @staticmethod
    def render(text, theme="592.2129", html=True, tab_length=0):
        """
        Renders input text to formatted text ready to send to the terminal.

        Args:
            text: input html text

        Returns:
            text formatted for print to terminal
        """

        if html:
            # Convert HTML
            parser = html2text.HTML2Text()
            parser.body_width = 0
            text = parser.handle(text)

            text = Query.escape(text)

        text = mdv.main(text, theme=theme, c_theme="953.3567", cols=180, tab_length=tab_length)

        if html:
            text = Query.unescape(text)

        return text.strip()

    @staticmethod
    def search(embeddings, cur, query, topn):
        """
        Executes an embeddings search for the input query. Each returned result is resolved
        to the full section row.

        Args:
            embeddings: embeddings model
            cur: database cursor
            query: query text
            topn: number of documents to return

        Returns:
            search results
        """

        if query == "*":
            return []

        results = []

        # Get list of required tokens
        must = [token.strip("+") for token in query.split() if token.startswith("+") and len(token) > 1]

        # Tokenize search query
        query = Tokenizer.tokenize(query)

        # Retrieve topn * 5 to account for duplicate matches
        for uid, score in embeddings.search(query, topn * 5):
            if score >= 0.6:
                cur.execute("SELECT Article, Text FROM sections WHERE id = ?", [uid])

                # Get matching row
                sid, text = cur.fetchone()

                # Add result if all required tokens are present or there are not required tokens
                if not must or all([token.lower() in text.lower() for token in must]):
                    # Save result
                    results.append((uid, score, sid, text))

        return results

    @staticmethod
    def highlights(results, topn):
        """
        Builds a list of highlights for the search results. Returns top ranked sections by importance
        over the result list.

        Args:
            results: search results
            topn: number of highlights to extract

        Returns:
            top ranked sections
        """

        sections = {}
        for uid, score, _, text in results:
            # Filter out lower scored results
            if score >= 0.35:
                sections[text] = (uid, text)

        # Return up to 5 highlights
        return Highlights.build(sections.values(), min(topn, 5))

    @staticmethod
    def documents(results, topn):
        """
        Processes search results and groups by article.

        Args:
            results: search results
            topn: number of documents to return

        Returns:
            results grouped by article
        """

        documents = {}

        # Group by article
        for _, score, article, text in results:
            if article not in documents:
                documents[article] = set()

            documents[article].add((score, text))

        # Sort based on section id, which preserves original order
        for uid in documents:
            documents[uid] = sorted(list(documents[uid]), reverse=True)

        # Get documents with top n best sections
        topn = sorted(documents, key=lambda k: max([x[0] for x in documents[k]]), reverse=True)[:topn]
        return {uid: documents[uid] for uid in topn}

    @staticmethod
    def all(cur):
        """
        Gets a list of all article ids.

        Args:
            cur: database cursor

        Returns:
            list of all ids as a dict
        """

        cur.execute("SELECT Id FROM articles LIMIT 500")
        return {row[0]: None for row in cur.fetchall()}

    @staticmethod
    def authors(authors):
        """
        Formats a short authors string

        Args:
            authors: full authors string

        Returns:
            short author string
        """

        if authors:
            authors = authors.split("; ")[0]
            if "," in authors:
                authors = authors.split(",")[0]
            else:
                authors = authors.split()[-1]

            return "%s et al" % authors

        return None

    @staticmethod
    def date(date):
        """
        Formats a date string.

        Args:
            date: input date string

        Returns:
            formatted date
        """

        if date:
            date = datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S")

            # 1/1 dates had no month/day specified, use only year
            if date.month == 1 and date.day == 1:
                return date.strftime("%Y")

            return date.strftime("%Y-%m-%d")

        return None

    @staticmethod
    def text(text):
        """
        Formats match text.

        Args:
            text: input text

        Returns:
            formatted text
        """

        if text:
            # Remove reference links ([1], [2], etc)
            text = re.sub(r"\s*[\[(][0-9, ]+[\])]\s*", " ", text)

            # Remove •
            text = text.replace("•", "")

            # Remove http links
            text = re.sub(r"http.+?\s", " ", text)

            # Remove boilerplate text
            text = re.sub(r"doi\s?:\s?(bioRxiv|medRxiv) preprint", " ", text)
            text = text.replace("All Rights Reserved", " ")

        return text

    @staticmethod
    def design(design):
        """
        Formats a study design field.

        Args:
            design: study design integer

        Returns:
            Study Design string
        """

        # Study design type mapping
        mapping = {1:"Systematic review", 2:"Randomized control trial", 3:"Non-randomized trial",
                   4:"Prospective observational", 5:"Time-to-event analysis", 6:"Retrospective observational",
                   7:"Cross-sectional", 8:"Case series", 9:"Modeling", 0:"Other"}

        return mapping[design]

    @staticmethod
    def sample(size, text):
        """
        Formats a sample string.

        Args:
            size: Sample size
            text: Sample text

        Returns:
            Formatted sample text
        """

        return "[%s] %s" % (size, Query.text(text)) if size else Query.text(text)

    @staticmethod
    def query(embeddings, db, query, topn):
        """
        Executes a query against the embeddings model.

        Args:
            embeddings: embeddings model
            db: open SQLite database
            query: query string
            topn: number of query results
        """

        # Default to 10 results if not specified
        topn = topn if topn else 10

        cur = db.cursor()

        print(Query.render("#Query: %s" % query, theme="729.8953") + "\n")

        # Query for best matches
        results = Query.search(embeddings, cur, query, topn)

        # Extract top sections as highlights
        print(Query.render("# Highlights"))
        for highlight in Query.highlights(results, int(topn / 5)):
            print(Query.render("## - %s" % Query.text(highlight)))

        print()

        # Get results grouped by document
        documents = Query.documents(results, topn)

        print(Query.render("# Articles") + "\n")

        # Print each result, sorted by max score descending
        for uid in sorted(documents, key=lambda k: sum([x[0] for x in documents[k]]), reverse=True):
            cur.execute("SELECT Title, Published, Publication, Design, Size, Sample, Method, Entry, Id, Reference FROM articles WHERE id = ?", [uid])
            article = cur.fetchone()

            print("Title: %s" % article[0])
            print("Published: %s" % Query.date(article[1]))
            print("Publication: %s" % article[2])
            print("Design: %s" % Query.design(article[3]))
            print("Sample: %s" % Query.sample(article[4], article[5]))
            print("Method: %s" % Query.text(article[6]))
            print("Entry: %s" % article[7])
            print("Id: %s" % article[8])
            print("Reference: %s" % article[9])

            # Print top matches
            for score, text in documents[uid]:
                print(Query.render("## - (%.4f): %s" % (score, Query.text(text)), html=False))

            print()

    @staticmethod
    def run(query, topn=None, path=None):
        """
        Executes a query against an index.

        Args:
            query: input query
            topn: number of results
            path: model path
        """

        # Load model
        embeddings, db = Models.load(path)

        # Query the database
        Query.query(embeddings, db, query, topn)

        # Free resources
        Models.close(db)

if __name__ == "__main__":
    if len(sys.argv) > 1:
        Query.run(sys.argv[1], int(sys.argv[2]) if len(sys.argv) > 2 else None, sys.argv[3] if len(sys.argv) > 3 else None)



"""
Config module
"""

import os

import yaml

class Task(object):
    """
    YAML task configuration loader
    """

    @staticmethod
    def load(task):
        """
        Loads a YAML configuration. Supports loading from a string or file path.

        Args:
            task: YAML string or file path to YAML configuration

        Returns:
            (name, queries, output directory)
        """

        if os.path.exists(task):
            # Load tasks yml file
            with open(task, "r") as f:
                # Read configuration
                config = yaml.safe_load(f)

                # Write output to same directory as input
                outdir = os.path.dirname(task)
        else:
            # Assume task is a yml string
            config = yaml.safe_load(task)
            outdir = "."

        # Extract queries
        queries = Task.queries(config)

        return config["name"], queries, outdir

    @staticmethod
    def queries(config):
        """
        Gets a list of queries from this configuration.

        Args:
            config: configuration object

        Returns:
            list of queries
        """

        queries = []

        for key, value in config.items():
            if key not in ["id", "name", "fields"]:
                # Flatten columns
                value["columns"] = Task.flatten(value["columns"])

                # Add query
                queries.append((key, value))

        return queries

    @staticmethod
    def flatten(columns):
        """
        Flattens nested lists into a single list.

        Args:
            columns: list of columns, possibly with nested lists

        Returns:
            flattened list of columns
        """

        output = []
        for column in columns:
            if isinstance(column, list):
                output.extend(column)
            else:
                output.append(column)

        return output



"""
Report factory module
"""

import os.path

from .csvr import CSV
from .markdown import Markdown
from .task import Task

from ..models import Models

class Execute(object):
    """
    Creates a Report
    """

    @staticmethod
    def create(render, embeddings, db, qa):
        """
        Factory method to construct a Report.

        Args:
            render: report rendering format
            embeddings: embeddings index
            db: database connection
            qa: qa model path

        Returns:
            Report
        """

        if render == "csv":
            return CSV(embeddings, db, qa)
        elif render == "md":
            return Markdown(embeddings, db, qa)

        return None

    @staticmethod
    def run(task, topn=None, render=None, path=None, qa=None):
        """
        Reads a list of queries from a task file and builds a report.

        Args:
            task: input task file
            topn: number of results
            render: report rendering format ("md" for markdown, "csv" for csv)
            path: embeddings model path
            qa: qa model path
        """

        # Load model
        embeddings, db = Models.load(path)

        # Read task configuration
        name, queries, outdir = Task.load(task)

        # Derive report format
        render = render if render else "md"

        # Create report object. Default to Markdown.
        report = Execute.create(render, embeddings, db, qa)

        # Generate output filename
        outfile = os.path.join(outdir, "%s.%s" % (name, render))

        # Stream report to file
        with open(outfile, "w") as output:
            # Build the report
            report.build(queries, topn, output)

        # Free any resources
        report.cleanup(outfile)

        # Free resources
        Models.close(db)



"""
Models module
"""

import os
import os.path
import sqlite3

from txtai.embeddings import Embeddings

class Models(object):
    """
    Common methods for generating data paths.
    """

    @staticmethod
    def basePath(create=False):
        """
        Base data path - ~/.cord19

        Args:
            create: if directory should be created

        Returns:
            path
        """

        # Get model cache path
        path = os.path.join(os.path.expanduser("~"), ".cord19")

        # Create directory if required
        if create:
            os.makedirs(path, exist_ok=True)

        return path

    @staticmethod
    def modelPath(create=False):
        """
        Model path for name

        Args:
            create: if directory should be created

        Returns:
            path
        """

        path = os.path.join(Models.basePath(), "models")

        # Create directory if required
        if create:
            os.makedirs(path, exist_ok=True)

        return path

    @staticmethod
    def vectorPath(name, create=False):
        """
        Vector path for name

        Args:
            name: vectors name
            create: if directory should be created

        Returns:
            path
        """

        path = os.path.join(Models.basePath(), "vectors")

        # Create directory path if required
        if create:
            os.makedirs(path, exist_ok=True)

        # Append file name to path
        return os.path.join(path, name)

    @staticmethod
    def load(path):
        """
        Loads an embeddings model and db database.

        Args:
            path: model path, if None uses default path

        Returns:
            (embeddings, db handle)
        """

        # Default path if not provided
        if not path:
            path = Models.modelPath()

        dbfile = os.path.join(path, "articles.sqlite")

        if os.path.isfile(os.path.join(path, "config")):
            print("Loading model from %s" % path)
            embeddings = Embeddings()
            embeddings.load(path)
        else:
            print("ERROR: loading model: ensure model is present")
            raise FileNotFoundError("Unable to load model from %s" % path)

        # Connect to database file
        db = sqlite3.connect(dbfile)

        return (embeddings, db)

    @staticmethod
    def close(db):
        """
        Closes a SQLite database database.

        Args:
            db: open database
        """

        # Free database resources
        db.close()



"""
paperai query shell module.
"""

import sys

from cmd import Cmd

from .models import Models
from .query import Query

class Shell(Cmd):
    """
    paperai query shell.
    """

    def __init__(self, path):
        super(Shell, self).__init__()

        self.intro = "paperai query shell"
        self.prompt = "(paperai) "

        self.embeddings = None
        self.db = None
        self.path = path

    def preloop(self):
        # Load embeddings and questions.db
        self.embeddings, self.db = Models.load(self.path)

    def postloop(self):
        Models.close(self.db)

    def default(self, line):
        Query.query(self.embeddings, self.db, line, None)

def main(path=None):
    """
    Shell execution loop.

    Args:
        path: model path
    """

    Shell(path).cmdloop()

if __name__ == "__main__":
    main(sys.argv[1] if len(sys.argv) > 1 else None)



"""
Defines main entry point for Report process.
"""

import sys

from .execute import Execute

if __name__ == "__main__":
    if len(sys.argv) > 1:
        # Run report with params: input file, topn, render format, embeddings model path, qa model path
        Execute.run(sys.argv[1],
                    int(sys.argv[2]) if len(sys.argv) > 2 else None,
                    sys.argv[3] if len(sys.argv) > 3 else None,
                    sys.argv[4] if len(sys.argv) > 4 else None,
                    sys.argv[5] if len(sys.argv) > 5 else None)



"""
CSV report module
"""

import csv
import os
import os.path

from ..query import Query

from .common import Report

class CSV(Report):
    """
    Report writer for CSV exports. Format is designed to be imported into other tools.
    """

    def __init__(self, embeddings, db, qa):
        super(CSV, self).__init__(embeddings, db, qa)

        # CSV writer handle
        self.csvout = None
        self.writer = None

    def cleanup(self, outfile):
        # Delete created master csv file
        os.remove(outfile)

    def query(self, output, task, query):
        # Close existing file
        if self.csvout:
            self.csvout.close()

        self.csvout = open(os.path.join(os.path.dirname(output.name), "%s.csv" % task), "w", newline="")
        self.writer = csv.writer(self.csvout, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL)

    def write(self, row):
        """
        Writes line to output file.

        Args:
            line: line to write
        """

        # Write csv line
        self.writer.writerow(row)

    def headers(self, columns, output):
        self.names = columns

        # Write out column names
        self.write(self.names)

    def buildRow(self, article, sections, calculated):
        columns = {}

        # Date
        columns["Date"] = Query.date(article[0]) if article[0] else ""

        # Study
        columns["Study"] = article[1]

        # Study Link
        columns["Study Link"] = article[2]

        # Journal
        columns["Journal"] = article[3] if article[3] else article[4]

        # Source
        columns["Source"] = article[4]

        # Study Type
        columns["Study Type"] = Query.design(article[5])

        # Sample Size
        columns["Sample Size"] = article[6]

        # Study Population
        columns["Study Population"] = Query.text(article[8] if article[8] else article[7])

        # Sample Text
        columns["Sample Text"] = article[7]

        # Top Matches
        columns["Matches"] = "\n\n".join([Query.text(text) for _, text in sections]) if sections else ""

        # Entry Date
        columns["Entry"] = article[9] if article[9] else ""

        # Merge in calculated fields
        columns.update(calculated)

        return columns

    def writeRow(self, output, row):
        self.write(row)



# pylint: disable = C0111
from setuptools import find_packages, setup

with open("README.md", "r") as f:
    DESCRIPTION = f.read()

setup(name="paperai",
      version="1.4.0",
      author="NeuML",
      description="AI-powered literature discovery and review engine for medical/scientific papers",
      long_description=DESCRIPTION,
      long_description_content_type="text/markdown",
      url="https://github.com/neuml/paperai",
      project_urls={
          "Documentation": "https://github.com/neuml/paperai",
          "Issue Tracker": "https://github.com/neuml/paperai/issues",
          "Source Code": "https://github.com/neuml/paperai",
      },
      license="Apache 2.0: http://www.apache.org/licenses/LICENSE-2.0",
      packages=find_packages(where="src/python"),
      package_dir={"": "src/python"},
      keywords="search embedding machine-learning nlp covid-19 medical scientific papers",
      python_requires=">=3.6",
      entry_points={
          "console_scripts": [
              "paperai = paperai.shell:main",
          ],
      },
      install_requires=[
          "html2text>=2020.1.16",
          "mdv>=1.7.4",
          "networkx>=2.4",
          "PyYAML>=5.3",
          "regex>=2020.5.14",
          "txtai>=1.4.0",
      ],
      classifiers=[
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Topic :: Scientific/Engineering :: Artificial Intelligence",
          "Topic :: Software Development",
          "Topic :: Text Processing :: Indexing",
          "Topic :: Utilities"
      ])



"""
Export module
"""

import os
import os.path
import sqlite3
import sys

import regex as re

# pylint: disable=E0611
# Defined at runtime
from .index import Index
from .models import Models

class Export(object):
    """
    Exports database rows into a text file line-by-line.
    """

    @staticmethod
    def stream(dbfile, output):
        """
        Iterates over each row in dbfile and writes text to output file

        Args:
            dbfile: SQLite file to read
            output: output file to store text
        """

        with open(output, "w") as output:
            # Connection to database file
            db = sqlite3.connect(dbfile)
            cur = db.cursor()

            # Get all indexed text, with a detected study design, excluding modeling designs
            cur.execute(Index.SECTION_QUERY + " AND design NOT IN (0, 9)")

            count = 0
            for _, name, text in cur:
                if not name or not re.search(Index.SECTION_FILTER, name.lower()):
                    count += 1
                    if count % 1000 == 0:
                        print("Streamed %d documents" % (count), end="\r")

                    # Write row
                    if text:
                        output.write(text + "\n")

            print("Iterated over %d total rows" % (count))

            # Free database resources
            db.close()

    @staticmethod
    def run(output, path):
        """
        Exports data from database to text file, line by line.

        Args:
            output: output file path
            path: model path, if None uses default path
        """

        # Default path if not provided
        if not path:
            path = Models.modelPath()

        # Derive path to dbfile
        dbfile = os.path.join(path, "articles.sqlite")

        # Stream text from database to file
        Export.stream(dbfile, output)

if __name__ == "__main__":
    # Export data
    Export.run(sys.argv[1], sys.argv[2] if len(sys.argv) > 2 else None)
